MODULE Mcs51XOH;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Host Interface
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
*)

IMPORT
  SYSTEM, Out, TextViews, TextModels, Files, Strings, Dialog, DevMarkers, Dates, Views, TextMappers, TextRulers, StdCmds;

CONST
(* Oberon constants *)
  MaxIdLen* = 32;  (* max. length of identifiers and keywords *)
  MaxStrLen* = 256;  (* max. length of strings *)

  ObjFileTag* = 0F1X;  (* to identify a Oberon object file *)
  SymFileTag* = 0F3X;  (* to identify a Oberon object file *)
  MarkTag* = 0FFFFH;  (* end of the fixup-chain of a module (in .obj-file) *)

  MaxNofImports* = 32;  (* max. number of imported modules *)
  MaxCodeSize* = 10000H;  (* max. code memory size (2^16) *)

(* File extensions *)
  ObjectFileExtension* = "obj";
  SymbolFileExtension* = "sym";
  BinaryFileExtension* = "bin";
  IntelHexFileExtension* = "hex";

(* ASCII constants *)
  TAB* = 09X;  (* horizontal tabulator *)
  CR* = 0DX;  (* carriage return *)
  LF* = 0AX;  (* line feed *)
  DoubleQuote* = 22X;  (* " *)

(* Host interface constants *)
  False = 0X; True = 1X;  (* for files *)

TYPE
  Id* = ARRAY MaxIdLen OF CHAR;  (* Type for identifiers *)
  Str* = ARRAY MaxStrLen OF CHAR;  (* Type for strings *)

  Window* = POINTER TO RECORD
    title: Views.Title;
    model: TextModels.Model;
    style: TextRulers.Style;
    formatter: TextMappers.Formatter
  END;

  InputFile* = POINTER TO RECORD
    name-: Files.Name;
    file: Files.File;
    reader: Files.Reader
  END;
  
  OutputFile* = POINTER TO RECORD
    name-: Files.Name;
    type: Files.Type;
    file: Files.File;
    writer: Files.Writer
  END;

VAR
  path*: ARRAY 32 OF CHAR;  (* default path, may be reset by the clients *)
  sourceReader: TextModels.Reader;  (* only one global source text reader *)
  sourceModel: TextModels.Model;  (* used by InsertError *)
  errorFile: InputFile;
  printerFile: OutputFile;
  
(* Source Window Reading *)

PROCEDURE SourceOpen*(): BOOLEAN;
VAR
  sourceView: TextViews.View;
BEGIN
  sourceModel := NIL; sourceReader := NIL; sourceView := TextViews.Focus();  (* currently selected window *)
  IF (sourceView # NIL) THEN sourceModel := sourceView.ThisModel() END;  (* model of that window *)
  IF (sourceModel # NIL) THEN sourceReader := sourceModel.NewReader(NIL) END;  (* reader of that model *)
  IF (sourceReader # NIL) THEN DevMarkers.Unmark(sourceModel); RETURN TRUE END;
  RETURN FALSE
END SourceOpen;

PROCEDURE SourceRead*(VAR ch: CHAR);
BEGIN IF (sourceReader # NIL) THEN sourceReader.ReadChar(ch) ELSE ch := 0X END
END SourceRead;

PROCEDURE SourcePos*(): LONGINT;
BEGIN IF (sourceReader # NIL) THEN RETURN sourceReader.Pos() ELSE RETURN -1 END
END SourcePos;

PROCEDURE SourceEot*(): BOOLEAN;
BEGIN  IF (sourceReader # NIL) THEN RETURN sourceReader.eot ELSE RETURN TRUE END
END SourceEot;

(* Text Writing in Log *)

PROCEDURE LogWrite*(ch: CHAR);
BEGIN Out.Char(ch)
END LogWrite;

PROCEDURE LogWriteString*(s: ARRAY OF CHAR);
BEGIN Out.String(s)
END LogWriteString;

PROCEDURE LogWriteInt*(i: LONGINT);
BEGIN Out.Int(i, 0)
END LogWriteInt;

PROCEDURE LogWriteHex8*(i: LONGINT);  (* 0 .. 0FFH *)
BEGIN
  IF (i < 0H) OR (i > 0FFH) THEN Out.String("--")
  ELSE
    IF (i DIV 10H > 9) THEN Out.Char(CHR(ORD("A") + (i DIV 10H) - 0AH))
    ELSE Out.Char(CHR(ORD("0") + (i DIV 10H)))
    END;
    IF (i MOD 10H > 9) THEN Out.Char(CHR(ORD("A") + (i MOD 10H) - 0AH))
    ELSE Out.Char(CHR(ORD("0") + (i MOD 10H)))
    END
  END
END LogWriteHex8;

PROCEDURE LogWriteHex16*(l: LONGINT);  (* 0 .. 0FFFFH *)
BEGIN
  IF (l < 0H) OR (l > 0FFFFH) THEN Out.String("----")
  ELSE LogWriteHex8(l DIV 100H); LogWriteHex8(l MOD 100H)
  END
END LogWriteHex16;

PROCEDURE LogWriteHex32*(l: LONGINT);
BEGIN LogWriteHex16(l DIV 10000H); LogWriteHex16(l MOD 10000H)
END LogWriteHex32;

PROCEDURE LogWriteLn*;
BEGIN Out.Ln;
END LogWriteLn;

(* Text Writing in Window *)

PROCEDURE NewWindow*(): Window;
VAR win: Window;
BEGIN
  NEW(win);  (* subject to be changed; 'win' shouldn't be created here (=> record extensions) *)
  win.model := TextModels.dir.New();
  win.formatter.ConnectTo(win.model);
  win.style := NIL;
  RETURN win
END NewWindow;

PROCEDURE (win: Window) SetRuler*(tabs: ARRAY OF LONGINT);
VAR
  i: LONGINT;
  prop: TextRulers.Prop;
  attr: TextRulers.Attributes;
BEGIN
  NEW(prop); i := 0;
  WHILE (i < LEN(tabs)) & (i < LEN(prop.tabs.tab)) DO prop.tabs.tab[i].val := tabs[i]; INC(i) END;
  prop.tabs.len := i; prop.valid := { TextRulers.tabs };
  win.style := TextRulers.dir.NewStyle(TextRulers.dir.attr);
  win.formatter.WriteView(TextRulers.dir.New(win.style));
  NEW(attr); attr.ModifyFrom(win.style.attr, prop); win.style.SetAttr(attr)
END SetRuler;

PROCEDURE (win: Window) Open*(title: ARRAY OF CHAR);
BEGIN COPY(title, win.title); Views.OpenAux(TextViews.dir.New(win.model), win.title)
END Open;

PROCEDURE (win: Window) Write*(ch: CHAR);
BEGIN win.formatter.WriteChar(ch)
END Write;

PROCEDURE (win: Window) WriteString*(str: ARRAY OF CHAR);
BEGIN win.formatter.WriteString(str)
END WriteString;

PROCEDURE (win: Window) WriteInt*(i: LONGINT);
BEGIN win.formatter.WriteInt(i)
END WriteInt;

PROCEDURE (win: Window) WriteHex8*(i: LONGINT);  (* 0 .. 0FFH *)
BEGIN
  IF (i < 0H) OR (i > 0FFH) THEN win.WriteString("--")
  ELSE
    IF (i DIV 10H > 9) THEN win.Write(CHR(ORD("A") + (i DIV 10H) - 0AH))
    ELSE win.Write(CHR(ORD("0") + (i DIV 10H)))
    END;
    IF (i MOD 10H > 9) THEN win.Write(CHR(ORD("A") + (i MOD 10H) - 0AH))
    ELSE win.Write(CHR(ORD("0") + (i MOD 10H)))
    END
  END
END WriteHex8;

PROCEDURE (win: Window) WriteHex16*(l: LONGINT);  (* 0 .. 0FFFFH *)
BEGIN
  IF (l < 0H) OR (l > 0FFFFH) THEN win.WriteString("----")
  ELSE win.WriteHex8(l DIV 100H); win.WriteHex8(l MOD 100H)
  END
END WriteHex16;

PROCEDURE (win: Window) WriteHex32*(l: LONGINT);
BEGIN win.WriteHex16(l DIV 10000H); win.WriteHex16(l MOD 10000H)
END WriteHex32;

PROCEDURE (win: Window) WriteTab*;
BEGIN win.formatter.WriteTab
END WriteTab;

PROCEDURE (win: Window) WriteLn*;
BEGIN win.formatter.WriteLn
END WriteLn;

(* File Reading *)

PROCEDURE NewInputFile*(path, name, type: ARRAY OF CHAR): InputFile;
VAR
  in: InputFile;
  loc: Files.Locator;
  tempName: Files.Name;
BEGIN
  NEW(in);  (* subject to be changed; 'in' shouldn't be created here (=> record extensions) *)
  COPY(name, tempName); Strings.Append(tempName, "."); Strings.Append(tempName, type);
  COPY(name, in.name); loc := Files.dir.This(""); loc := loc.This(path);
  IF (loc # NIL) THEN in.file := Files.dir.Old(loc, tempName, Files.shared) ELSE RETURN NIL END;
  IF (in.file # NIL) THEN in.reader := in.file.NewReader(NIL) ELSE RETURN NIL END;
  IF (in.reader # NIL) THEN in.reader.SetPos(0) ELSE RETURN NIL END;  (* start reading at the beginning *)
  RETURN in
END NewInputFile;

PROCEDURE SelectInputFile*(path, defaultType: ARRAY OF CHAR): InputFile;  (* opens a file selector box *)
VAR
  in: InputFile;
  name: Files.Name;
  type: Files.Type;
  loc: Files.Locator;
  i: INTEGER;
BEGIN
  COPY(defaultType, type); loc := Files.dir.This(""); loc := loc.This(path); Dialog.GetIntSpec(type, loc, name);
  NEW(in);  (* subject to be changed; 'in' shouldn't be created here (=> record extensions) *)
  i := 0; WHILE (name[i] # 0X) & (name[i] # ".") DO in.name[i] := name[i]; INC(i) END; in.name[i] := 0X;
  IF (loc # NIL) THEN in.file := Files.dir.Old(loc, name, Files.shared) ELSE RETURN NIL END;
  IF (in.file # NIL) THEN in.reader := in.file.NewReader(NIL) ELSE RETURN NIL END;
  IF (in.reader # NIL) THEN in.reader.SetPos(0) ELSE RETURN NIL END;  (* start reading at the beginning *)
  RETURN in
END SelectInputFile;

PROCEDURE InputFileFrom*(out: OutputFile): InputFile;  (* to be able to read from an output file *)
VAR in: InputFile;
BEGIN
  NEW(in);  (* subject to be changed; 'in' shouldn't be created here (=> record extensions) *)
  COPY(out.name, in.name); in.file := out.file;
  IF (in.file # NIL) THEN in.reader := in.file.NewReader(NIL) ELSE RETURN NIL END;
  IF (in.reader # NIL) THEN in.reader.SetPos(0) ELSE RETURN NIL END;  (* start reading at the beginning *)
  RETURN in
END InputFileFrom;

PROCEDURE (in: InputFile) ReadBool*(VAR b: BOOLEAN);  (* TRUE, FALSE *)
VAR ch: CHAR;
BEGIN in.reader.ReadByte(ch); b := (ch = True)
END ReadBool;

PROCEDURE (in: InputFile) ReadSInt*(VAR s: SHORTINT);  (* -80H .. 7FH *)
VAR ch: CHAR;
BEGIN
  in.reader.ReadByte(ch);
  IF (ORD(ch) > 7FH) THEN s := SHORT(ORD(ch)-256)
  ELSE s := SHORT(ORD(ch)) END
END ReadSInt;

PROCEDURE (in: InputFile) ReadInt*(VAR i: INTEGER);  (* -8000H .. 7FFFH *)
VAR ch: CHAR;
BEGIN
  in.reader.ReadByte(ch); i := ORD(ch);  (* LSB *)
  in.reader.ReadByte(ch); i := i + 256 * ORD(ch)  (* MSB *)
END ReadInt;

PROCEDURE (in: InputFile) ReadLInt*(VAR l: LONGINT);  (* -80000000H .. 7FFFFFFFH *)
VAR ch: CHAR;
BEGIN
  in.reader.ReadByte(ch); l := ORD(ch);  (* LSB *)
  in.reader.ReadByte(ch); l := l + 256 * LONG(ORD(ch));  (* 'LONG' is necessary ! *)
  in.reader.ReadByte(ch); l := l + 256*256 * LONG(ORD(ch));
  in.reader.ReadByte(ch); l := l + 256*256*256 * LONG(ORD(ch))  (* MSB *)
END ReadLInt;

PROCEDURE (in: InputFile) Read*(VAR ch: CHAR);
BEGIN in.reader.ReadByte(ch)
END Read;

PROCEDURE (in: InputFile) ReadHex8*(VAR i: LONGINT);  (* 0 .. 0FFH *)
VAR ch: CHAR;
BEGIN
  in.reader.ReadByte(ch);
  IF ("0" <= ch) & (ch <= "9") THEN i := ORD(ch) - ORD("0") ELSE i := ORD(ch) - ORD("A") + 0AH END;
  in.reader.ReadByte(ch);
  IF ("0" <= ch) & (ch <= "9") THEN i := 10H*i + ORD(ch) - ORD("0") ELSE i := 10H*i + ORD(ch) - ORD("A") + 0AH END;
END ReadHex8;

PROCEDURE (in: InputFile) ReadWord*(VAR l: LONGINT);  (* 0 .. 0FFFFH *)
VAR ch: CHAR;
BEGIN
  in.reader.ReadByte(ch); l := ORD(ch);  (* LSB *)
  in.reader.ReadByte(ch); l := l + 256 * LONG(ORD(ch))  (* MSB *)
END ReadWord;

PROCEDURE (in: InputFile) ReadBytes*(VAR bytes: ARRAY OF CHAR; nofBytes: LONGINT);
BEGIN in.reader.ReadBytes(bytes, 0, nofBytes)  (* 0: start writing in 'bytes' at position 0 *)
END ReadBytes;

PROCEDURE (in: InputFile) ReadString*(VAR str: ARRAY OF CHAR);
VAR ch: CHAR; i: LONGINT;
BEGIN i := 0; REPEAT in.reader.ReadByte(ch); str[i] := ch; INC(i) UNTIL (i = LEN(str)) OR (ch = 0X)
END ReadString;

PROCEDURE (in: InputFile) Eof*(): BOOLEAN;
BEGIN RETURN in.reader.eof;
END Eof;

PROCEDURE (in: InputFile) Pos*(): LONGINT;
BEGIN RETURN in.reader.Pos()
END Pos;

PROCEDURE (in: InputFile) SetPos*(pos: LONGINT);
BEGIN in.reader.SetPos(pos)
END SetPos;

PROCEDURE (in: InputFile) Close*;
BEGIN in.file.Close
END Close;

(* File Writing *)

PROCEDURE NewOutputFile*(path, name, type: ARRAY OF CHAR): OutputFile;
VAR
  out: OutputFile;
  loc: Files.Locator;
BEGIN
  NEW(out);  (* subject to be changed; 'out' shouldn't be created here (=> record extensions) *)
  COPY(name, out.name); COPY(type, out.type); loc := Files.dir.This(""); loc := loc.This(path);
  IF (loc # NIL) THEN out.file := Files.dir.New(loc); out.writer := out.file.NewWriter(NIL) ELSE RETURN NIL END;
  IF (out.writer # NIL) THEN out.writer.SetPos(0) ELSE RETURN NIL END;  (* start writing at the beginning *)
  RETURN out
END NewOutputFile;

PROCEDURE (out: OutputFile) WriteBool*(b: BOOLEAN);  (* TRUE, FALSE *)
BEGIN IF b THEN out.writer.WriteByte(True) ELSE out.writer.WriteByte(False) END
END WriteBool;

PROCEDURE (out: OutputFile) WriteSInt*(s: SHORTINT);  (* -80H .. 7FH *)
BEGIN
  IF (s < 0) THEN out.writer.WriteByte(CHR(s+256))
  ELSE out.writer.WriteByte(CHR(s)) END
END WriteSInt;

PROCEDURE (out: OutputFile) WriteInt*(i: INTEGER);  (* -8000H .. 7FFFH *)
BEGIN
  out.writer.WriteByte(CHR(i MOD 256));  (* LSB *)
  out.writer.WriteByte(CHR(i DIV 256))  (* MSB *)
END WriteInt;

PROCEDURE (out: OutputFile) WriteLInt*(l: LONGINT);  (* -80000000H .. 7FFFFFFFH *)
BEGIN
  out.writer.WriteByte(CHR(l MOD 256));  (* LSB *)
  l := l DIV 256; out.writer.WriteByte(CHR(l MOD 256));
  l := l DIV 256; out.writer.WriteByte(CHR(l MOD 256));
  l := l DIV 256; out.writer.WriteByte(CHR(l MOD 256))  (* MSB *)
END WriteLInt;

PROCEDURE (out: OutputFile) Write*(ch: CHAR);
BEGIN out.writer.WriteByte(ch)
END Write;

PROCEDURE (out: OutputFile) WriteHex8*(i: LONGINT);  (* 0 .. 0FFH *)
BEGIN
  IF (i < 0H) OR (i > 0FFH) THEN out.writer.WriteByte("-"); out.writer.WriteByte("-")
  ELSE
    IF (i DIV 10H > 9) THEN out.writer.WriteByte(CHR(ORD("A") + (i DIV 10H) - 0AH))
    ELSE out.writer.WriteByte(CHR(ORD("0") + (i DIV 10H)))
    END;
    IF (i MOD 10H > 9) THEN out.writer.WriteByte(CHR(ORD("A") + (i MOD 10H) - 0AH))
    ELSE out.writer.WriteByte(CHR(ORD("0") + (i MOD 10H)))
    END
  END
END WriteHex8;

PROCEDURE (out: OutputFile) WriteWord*(l: LONGINT);  (* 0 .. 0FFFFH *)
BEGIN
  out.writer.WriteByte(CHR(l MOD 256));  (* LSB *)
  out.writer.WriteByte(CHR(l DIV 256))  (* MSB *)
END WriteWord;

PROCEDURE (out: OutputFile) WriteBytes*(VAR bytes: ARRAY OF CHAR; nofBytes: LONGINT);
BEGIN out.writer.WriteBytes(bytes, 0, nofBytes)  (* 0: start reading from 'bytes' at position 0 *)
END WriteBytes;

PROCEDURE (out: OutputFile) WriteString*(str: ARRAY OF CHAR);
VAR i: LONGINT;
BEGIN i := 0; REPEAT out.writer.WriteByte(str[i]); INC(i) UNTIL (str[i-1] = 0X)
END WriteString;

PROCEDURE (out: OutputFile) Pos*(): LONGINT;
BEGIN RETURN out.writer.Pos()
END Pos;

PROCEDURE (out: OutputFile) SetPos*(pos: LONGINT);
BEGIN out.writer.SetPos(pos)
END SetPos;

PROCEDURE (out: OutputFile) Register*;  (* the file will only be stored to disk if 'Register' is called *)
VAR res: LONGINT;
BEGIN out.file.Flush; out.file.Register(out.name, out.type, res)
END Register;

PROCEDURE (out: OutputFile) Close*;
BEGIN out.file.Close
END Close;

(* Parallel Port Access *)

PROCEDURE [code] OutLPT*(port: INTEGER; ch: CHAR)  (* 'port' = 0 for LPT1:, = 1 for LPT2: and so on *)
  (* the first parameter ('port') is in EAX, the remaining parameters are pushed *)
  (* onto the stack from right to left. The return value must be in EAX *)
  89H, 0C2H,  (* MOV DX, AX *)  (* 'port' is now in DX which selects the port to be used *)
  58H,  (* POP AX *)  (* 'ch' is now in AL which selects the character to be sent *)
  0B4H, 0H,  (* MOV AH, 0H *)  (* send character *)
  0CDH, 17H;  (* INT 17H *)  (* execute *)
  
PROCEDURE [code] InitLPT*(port: INTEGER): BOOLEAN  (* returns FALSE if failed *)
  89H, 0C2H,  (* MOV DX, AX *)  (* 'port' is now in DX which selects the port to be used *)
  0B4H, 01H,  (* MOV AH, 1H *)  (* init printer port *)
  0CDH, 17H,  (* INT 17H *)  (* execute *)
  80H, 0E4H, 80H,  (* AND AH, 80H *)  (* mask 'not ready bit' *)
  0D0H, 0C4H,  (* ROL AH,  1 *)  (* TRUE = 1 *)
  88H, 0E0H;  (* MOV AL, AH *)  (* return value in AL *)

(* Miscellaneous *)

PROCEDURE InsertError*(nr: INTEGER);
VAR
  i, j: INTEGER;
  str, errNr: ARRAY 16 OF CHAR;
  msg: Str;
  marker: DevMarkers.View;
BEGIN
  marker := DevMarkers.dir.New(DevMarkers.mark);
  marker.InitErr(nr);
  j := 0;
  REPEAT errNr[j] := CHR(nr MOD 10 + ORD("0")); INC(j); nr := nr DIV 10
  UNTIL (nr = 0);
  str := "#Mcs51:"; i := 7;
  REPEAT DEC(j); str[i] := errNr[j]; INC(i) UNTIL (j = 0);
  str[i] := 0X;
  Dialog.MapString(str, msg); marker.InitMsg(msg);
  DevMarkers.Insert(sourceModel, sourceReader.Pos() - 1, marker)
END InsertError;

PROCEDURE ShowFirstError*;
BEGIN DevMarkers.ShowFirstError(sourceModel, FALSE)  (* last parameter: focusOnly *)
END ShowFirstError;

PROCEDURE GetClock*(VAR time, date: LONGINT);
VAR
  curTime: Dates.Time;
  curDate: Dates.Date;
BEGIN
  (* hour: 0 .. 23; minute: 0 .. 59; second: 0 .. 59; => time: 0 .. 86399 *)
  Dates.GetTime(curTime); time := curTime.second + 60*curTime.minute + 60*60*curTime.hour;
  (* year: 1 .. 9999; month: 1 .. 12; day: 1 .. 31; => date: 0 .. 3719627 *)
  Dates.GetDate(curDate); date := (curDate.day-1) + 31*(curDate.month-1) + 12*31*(curDate.year-1)
END GetClock;

PROCEDURE OpenToolDialog*(fileName, dialogTitle: ARRAY OF CHAR);
BEGIN StdCmds.OpenToolDialog(fileName, dialogTitle)
END OpenToolDialog;

PROCEDURE CloseDialog*;
BEGIN StdCmds.CloseDialog
END CloseDialog;

BEGIN path := "Mcs51\Mod"
END Mcs51XOH.

