MODULE Mcs51XOL;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Linker and Loader
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
Based on    : NW 16.02.86 / 07.04.91
*)

IMPORT
  XOH := Mcs51XOH, XOC := Mcs51XOC;

CONST
  DefaultSp = 7H;  (* value of the stack pointer after reset *)
  IntelHexNofData = 10H;  (* intel-hex files usually have 16 data bytes each line *)

(* Result values *)
  Ok = 0;  (* linking complete *)
  ObjNotFound = 1;  (* object file not available *)
  BadObjFile = 2;  (* referenced file is not an object file *)
  BadKey = 3;  (* key mismatch *)
  OptionsInconsistency = 4;  (* options inconsistency *)
  MultiplyInterrupt = 5;  (* multiply defined interrupt service procedure *)
  MultiplyTrap = 6;  (* multiply defined trap procedure *)
  CmdNotFound = 7;  (* command not found *)
  CodeTooLong = 8;  (* linked code too long *)
  NoOutputFile = 9;  (* output file could not be generated *)

TYPE
  FixupLocation = POINTER TO RECORD  (* elements of the fixup lists *)
    pc: LONGINT;
    entryNo: LONGINT;  (* only relevant in the 'importedProcCallList' *)
    next: FixupLocation
  END;

  Command = RECORD
    name: XOH.Id;
    entryAdr: LONGINT
  END;

  Module = POINTER TO RECORD
    name: XOH.Id;
    key: LONGINT;
    nofImports, nofEntries, nofCommands, constantsSize, globalData, maxLocalData, codeSize: LONGINT;
    importTab: POINTER TO ARRAY OF Module;  (* will be allocated dynamically *)
    entryTab: POINTER TO ARRAY OF LONGINT;
    commandTab: POINTER TO ARRAY OF Command;
    constantsTab: POINTER TO ARRAY OF CHAR;
    code: POINTER TO ARRAY OF CHAR;
    codeBase: LONGINT;  (* start address of the code of that module in the final layout *)
    globalBase: LONGINT;  (* base address of its global variables *)
    globalVarAccessList, localVarAccessList, addressAccessList, trapCallList: FixupLocation;
    importedProcCallList: ARRAY XOH.MaxNofImports OF FixupLocation;
    importedVarAccessList: ARRAY XOH.MaxNofImports OF FixupLocation;
    next: Module
  END;

  InterruptProcedure = POINTER TO RECORD
    intAdr: LONGINT;  (* address where the microcontroller branches to in case of that interrupt *)
    intProcAdr: LONGINT;  (* entry point of the interrupt service procedure *)
    owner: Module;  (* module where the interrupt service procedure is defined *)
                (* must be known to get 'codeBase' of that module *)
    next: InterruptProcedure
  END;

VAR
(* Options (will be set by module 'MicroOberon') *)
  param*: XOH.Str;  (* textual parameter for linker *)
  intelHex*: BOOLEAN;  (* work with intel-hex files if TRUE; with binary files otherwise *)
  lptPort*: INTEGER;  (* LPT port for downloader. = 0 for LPT1:, =1 for LPT2:, =2 for LPT3: and so on *)

  result: SHORTINT;  (* global error number for the linker *)
  importing, imported: XOH.Id;  (* to display the key inconsistency relation *)
  pc, globalData, maxLocalData: LONGINT;  (* these three numbers are displayed at the end of the linking process *)
                                (* 'globalData' and 'maxLocalData' are valid after loading of the modules *)
  code: ARRAY XOH.MaxCodeSize OF CHAR;
  sp: LONGINT;  (* after 'CreateLayout', 'sp' contains the initial value of the stack pointer (for 'DisplayInformation') *)

  options: RECORD  (* these options must be equal in all modules to be linked *)
    valid: BOOLEAN;  (* will be set to TRUE after 'options' is set for the first time *)
    localDataWindow: LONGINT;  (* max. size of local data of procedures *)
    intRAM: LONGINT;  (* upper bound of the internal RAM (128 or 256 bytes respectively) *)
    extRAMbeg: LONGINT;  (* lower bound of the external RAM (up to 65536) *)
    extRAMend: LONGINT  (* upper bound of the external RAM (up to 65536) *)
  END;

  moduleList: Module;  (* list of all loaded modules *)
  intProcList: InterruptProcedure;  (* list of all interrupt procedures *)
  trapProc: InterruptProcedure;  (* the trap procedure is an interrupt procedure with 'intAdr' = 0 *)

(* Linker *)

PROCEDURE ThisModule(VAR name: ARRAY OF CHAR): Module;  (* load a module *)
VAR
  objFile: XOH.InputFile;
  ch: CHAR;
  i: INTEGER;
  l, impKey: LONGINT;
  mod, impMod: Module;
  impName: XOH.Id;
  intProc: InterruptProcedure;
  fix: FixupLocation;
BEGIN
  result := Ok; mod := moduleList;  (* mod = first module in list *)
  WHILE (mod # NIL) & (name # mod.name) DO mod := mod.next END;
  IF (mod # NIL) THEN RETURN mod END;  (* module already loaded *)
  objFile := XOH.NewInputFile(XOH.path, name, XOH.ObjectFileExtension);
  IF (objFile = NIL) THEN COPY(name, imported); result := ObjNotFound; RETURN NIL END;
  objFile.Read(ch);  (* read object file tag *)
  IF (ch # XOH.ObjFileTag) THEN result := BadObjFile; objFile.Close; RETURN NIL END;
  NEW(mod);  (* create new module object *)
  mod.globalVarAccessList := NIL; mod.localVarAccessList := NIL; mod.addressAccessList := NIL; mod.trapCallList := NIL;
  i := 0; WHILE (i < XOH.MaxNofImports) DO
    mod.importedProcCallList[i] := NIL; mod.importedVarAccessList[i] := NIL; INC(i)
  END;  (* empty all lists *)
(* HeaderBlock *)
  objFile.ReadWord(mod.nofImports); objFile.ReadWord(mod.nofEntries); objFile.ReadWord(mod.nofCommands);
  objFile.ReadWord(mod.constantsSize); objFile.ReadWord(mod.globalData); objFile.ReadWord(mod.maxLocalData);
  objFile.ReadWord(mod.codeSize); objFile.ReadLInt(mod.key); objFile.ReadString(mod.name);
(* OptionsBlock *)
  IF ~options.valid THEN  (* read options for the first time *)
    options.valid := TRUE;
    objFile.ReadWord(options.localDataWindow); objFile.ReadWord(options.intRAM);
    objFile.ReadWord(options.extRAMbeg); objFile.ReadWord(options.extRAMend)
  ELSE  (* compare with existing options *)
    objFile.ReadWord(l); IF (l # options.localDataWindow) THEN result := OptionsInconsistency END;
    objFile.ReadWord(l); IF (l # options.intRAM) THEN result := OptionsInconsistency END;
    objFile.ReadWord(l); IF (l # options.extRAMbeg) THEN result := OptionsInconsistency END;
    objFile.ReadWord(l); IF (l # options.extRAMend) THEN result := OptionsInconsistency END
  END;
  IF (result # Ok) THEN objFile.Close; RETURN NIL END;
  IF (mod.nofImports > 0) THEN NEW(mod.importTab, mod.nofImports) END;
  IF (mod.nofEntries > 0) THEN NEW(mod.entryTab, mod.nofEntries) END;
  IF (mod.nofCommands > 0) THEN NEW(mod.commandTab, mod.nofCommands) END;
  IF (mod.constantsSize > 0) THEN NEW(mod.constantsTab, mod.constantsSize) END;
  IF (mod.codeSize > 0) THEN NEW(mod.code, mod.codeSize) END;
  INC(globalData, mod.globalData);  (* 'globalData' is valid after the loading of all modules *)
  IF (mod.maxLocalData > maxLocalData) THEN maxLocalData := mod.maxLocalData END;  (* adjust the global maximum *)
(* ImportBlock *)
  i := 0;
  WHILE (i < mod.nofImports) & ~objFile.Eof() DO
    objFile.ReadString(impName); objFile.ReadLInt(impKey);
    impMod := ThisModule(impName);
    IF (result # Ok) THEN objFile.Close; RETURN NIL
    ELSE
      IF (impMod.key = impKey) THEN mod.importTab[i] := impMod; INC(i)
      ELSE COPY(name, importing); COPY(impName, imported); result := BadKey; objFile.Close; RETURN NIL
      END
    END
  END;
(* EntryBlock *)
  i := 0;
  WHILE (i < mod.nofEntries) & ~objFile.Eof() DO
    objFile.ReadWord(mod.entryTab[i]); INC(i)
  END;
(* CommandBlock *)
  i := 0;
  WHILE (i < mod.nofCommands) & ~objFile.Eof() DO
    objFile.ReadString(mod.commandTab[i].name);
    objFile.ReadWord(mod.commandTab[i].entryAdr); INC(i)
  END;
(* TrapProcedure & InterruptBlock *)
  objFile.ReadWord(l);
  WHILE (l # XOH.MarkTag) & ~objFile.Eof() DO
    NEW(intProc); intProc.intAdr := l; intProc.owner := mod;
    objFile.ReadWord(intProc.intProcAdr);
    IF (intProc.intAdr = 0) THEN  (* trap procedure *)
      IF (trapProc = NIL) THEN intProc.next := NIL; trapProc := intProc
      ELSE result := MultiplyTrap; objFile.Close; RETURN NIL  (* multiply defined trap procedure *)
      END
    ELSE  (* interrupt procedure; insert into global list *)
      intProc.next := intProcList; intProcList := intProc
    END;
    objFile.ReadWord(l)
  END;
(* ConstantsBlock *)
  IF (mod.constantsSize > 0) THEN objFile.ReadBytes(mod.constantsTab^, mod.constantsSize) END;
(* CodeBlock *)
  IF (mod.codeSize > 0) THEN objFile.ReadBytes(mod.code^, mod.codeSize) END;
(* ImportAccesses *)
  i := 0;
  WHILE (i < mod.nofImports) & ~objFile.Eof() DO
    objFile.ReadWord(l);  (* entryNo:2 *)
    WHILE (l # XOH.MarkTag) & ~objFile.Eof() DO
      NEW(fix); fix.entryNo := l; objFile.ReadWord(fix.pc); objFile.ReadWord(l);
      fix.next := mod.importedProcCallList[i]; mod.importedProcCallList[i] := fix
    END;
    objFile.ReadWord(l);  (* pc:2 *)
    WHILE (l # XOH.MarkTag) & ~objFile.Eof() DO
      NEW(fix); fix.pc := l; objFile.ReadWord(l);
      fix.next := mod.importedVarAccessList[i]; mod.importedVarAccessList[i] := fix
    END;
    INC(i)
  END;
(* GlobalVarAccesses *)
  objFile.ReadWord(l);
  WHILE (l # XOH.MarkTag) & ~objFile.Eof() DO
    NEW(fix); fix.pc := l; objFile.ReadWord(l);
    fix.next := mod.globalVarAccessList; mod.globalVarAccessList := fix
  END;
(* LocalVarAccesses *)
  objFile.ReadWord(l);
  WHILE (l # XOH.MarkTag) & ~objFile.Eof() DO
    NEW(fix); fix.pc := l; objFile.ReadWord(l);
    fix.next := mod.localVarAccessList; mod.localVarAccessList := fix
  END;
(* AddressAccesses *)
  objFile.ReadWord(l);
  WHILE (l # XOH.MarkTag) & ~objFile.Eof() DO
    NEW(fix); fix.pc := l; objFile.ReadWord(l);
    fix.next := mod.addressAccessList; mod.addressAccessList := fix
  END;
(* TrapCalls *)
  objFile.ReadWord(l);
  WHILE (l # XOH.MarkTag) & ~objFile.Eof() DO
    NEW(fix); fix.pc := l; objFile.ReadWord(l);
    fix.next := mod.trapCallList; mod.trapCallList := fix
  END;
(* Insert the loaded module at the end of the module list *)
  IF (moduleList = NIL) THEN moduleList := mod
  ELSE
    impMod := moduleList;  (* 'impMod' serves hier as current pointer *)
    WHILE (impMod.next # NIL) DO impMod := impMod.next END;
    impMod.next := mod;
  END;
  objFile.Close; result := Ok; RETURN mod
END ThisModule;

PROCEDURE MOVdc(adr, const: LONGINT);  (* insert a 'MOV direct,#data' instruction *)
BEGIN code[pc] := CHR(XOC.MOVdc); code[pc+1] := CHR(adr); code[pc+2] := CHR(const); INC(pc, 3)
END MOVdc;

PROCEDURE CreateLayout(mainMod: Module; cmdName: ARRAY OF CHAR);  (* is called after having loaded all modules *)
VAR
  curIntProc: InterruptProcedure;
  curMod: Module;
  curFix: FixupLocation;
  l, l0, l1, addr, codeOrigin, cmdAdr: LONGINT;
  intRamPtr: LONGINT;  (* will be incremented while assigning addresses for global variables *)
BEGIN
  result := Ok;
(* Init memory *)
  pc := 0; codeOrigin := 0;  (* if no interrupt procedure is defined, the code shall start at address 0 *)
  IF (intProcList # NIL) THEN  (* search for the interrupt with the highest address and insert a jump over it *)
    curIntProc := intProcList;
    WHILE (curIntProc # NIL) DO
      IF (curIntProc.intAdr > codeOrigin) THEN codeOrigin := curIntProc.intAdr END;
      curIntProc := curIntProc.next
    END;
    INC(codeOrigin, 3);  (* the code starts directly after the last 'LJMP' to the interrupt service procedure *)
    IF (codeOrigin - 2 < 7FH) THEN  (* SJMP *)
      code[pc] := CHR(XOC.SJMP);  (* if interrupts are defined, the controller must jump over them after reset *)
      code[pc+1] := CHR(codeOrigin - 2);  (* relative to the address of the next instruction *)
      INC(pc, 2)
    ELSE  (* LJMP *)
      code[pc] := CHR(XOC.LJMP);
      code[pc+1] := CHR(codeOrigin DIV 100H);
      code[pc+2] := CHR(codeOrigin MOD 100H);
      INC(pc, 3)
    END
  END;
  WHILE (pc < codeOrigin) DO code[pc] := 0X; INC(pc) END;  (* fill up this section with 0X = 'NOP' *)
  WHILE (pc < XOH.MaxCodeSize) DO code[pc] := 0FFX; INC(pc) END;  (* 0FFX: like an empty EPROM *)
(* Init the stack pointer and the heap pointer *)
  pc := codeOrigin;
  IF (options.extRAMbeg = options.extRAMend) THEN intRamPtr := XOC.NofRegs  (* no external RAM available *)
  ELSE intRamPtr := XOC.NofRegs + XOC.NofExtPtr  (* heap pointer used *)
  END;
  sp := intRamPtr + globalData + maxLocalData - 1;  (* the stack pointer will be preincremented on a PUSH *)
    (* 'globalData' here contains the cumulated sum of all module's global data *)
    (* 'maxLocalData' here contains the global maximum of all module *)
  IF (sp # DefaultSp) THEN MOVdc(XOC.SP, sp) END;
  IF (options.extRAMbeg # options.extRAMend) THEN  (* init heap pointer *)
    MOVdc(8H, options.extRAMbeg MOD 100H);  (* LSB of extRAMbeg *)
    MOVdc(9H, options.extRAMbeg DIV 100H);  (* MSB of extRAMbeg *)
  END;
(* Insert the code and fixup the absolute 16 bit addresses *)
  curMod := moduleList;
  WHILE (curMod # NIL) DO
    curMod.codeBase := pc;  (* code base address of that module *)
    l := 0;
    WHILE (l < curMod.codeSize) DO  (* copy code *)
      code[pc] := curMod.code[l]; INC(pc); INC(l);
      IF (pc = XOH.MaxCodeSize) THEN result := CodeTooLong; RETURN END
    END;
    curFix := curMod.addressAccessList;
    WHILE (curFix # NIL) DO
      addr := ORD(code[curMod.codeBase + curFix.pc]) * 100H + ORD(code[curMod.codeBase + curFix.pc + 1]);
      INC(addr, curMod.codeBase);
      code[curMod.codeBase + curFix.pc] := CHR(addr DIV 100H);
      code[curMod.codeBase + curFix.pc + 1] := CHR(addr MOD 100H);
      curFix := curFix.next
    END;
    curMod := curMod.next
  END;
(* Insert 'LJMP' instructions at the used interrupt addresses to the corresponding interrupt service procedures *)
  curIntProc := intProcList;
  WHILE (curIntProc # NIL) DO
    addr := curIntProc.owner.codeBase + curIntProc.intProcAdr;
    IF (code[curIntProc.intAdr] # CHR(XOC.NOP)) THEN  (* multiply defined interrupt address *)
      result := MultiplyInterrupt; RETURN
    ELSE
      code[curIntProc.intAdr] := CHR(XOC.LJMP);
      code[curIntProc.intAdr+1] := CHR(addr DIV 100H);
      code[curIntProc.intAdr+2] := CHR(addr MOD 100H)
    END;
    curIntProc := curIntProc.next
  END;
(* Fixup all references to imported procedures *)
  curMod := moduleList;
  WHILE (curMod # NIL) DO
    l0 := 0;
    WHILE (l0 < curMod.nofImports) DO  (* iterate over all modules imported by 'curMod' *)
      curFix := curMod.importedProcCallList[l0];
      WHILE (curFix # NIL) DO
        addr := curMod.importTab[l0].codeBase + curMod.importTab[l0].entryTab[curFix.entryNo];  (* effective address *)
        l := curFix.pc;
        WHILE (l # 0) DO  (* FixupChain16 *)
          l1 := ORD(code[curMod.codeBase + l]) * 100H + ORD(code[curMod.codeBase + l + 1]);  (* next location *)
          code[curMod.codeBase + l] := CHR(addr DIV 100H);
          code[curMod.codeBase + l + 1] := CHR(addr MOD 100H);
          l := l1  (* go to next location *)
        END;
        curFix := curFix.next
      END;
      INC(l0)
    END;
    curMod := curMod.next
  END;
(* Fixup 8 bit addresses of global variables *)
  curMod := moduleList;
  WHILE (curMod # NIL) DO
    curFix := curMod.globalVarAccessList;
    WHILE (curFix # NIL) DO
      l := ORD(code[curMod.codeBase + curFix.pc]); INC(l, intRamPtr);
      code[curMod.codeBase + curFix.pc] := CHR(l);
      curFix := curFix.next
    END;
    curMod.globalBase := intRamPtr;
    INC(intRamPtr, curMod.globalData);
    curMod := curMod.next
  END;
(* Fixup 8 bit addresses of local variables; 'intRamPtr' is here the start address of the local variables window *)
  curMod := moduleList;
  WHILE (curMod # NIL) DO
    curFix := curMod.localVarAccessList;
    WHILE (curFix # NIL) DO
      l := ORD(code[curMod.codeBase + curFix.pc]); INC(l, intRamPtr);
      code[curMod.codeBase + curFix.pc] := CHR(l);
      curFix := curFix.next
    END;
    curMod := curMod.next
  END;
(* Fixup 8 bit addresses of imported variables *)
  curMod := moduleList;
  WHILE (curMod # NIL) DO
    l0 := 0;
    WHILE (l0 < curMod.nofImports) DO  (* iterate over all modules imported by 'curMod' *)
      curFix := curMod.importedVarAccessList[l0];
      WHILE (curFix # NIL) DO
        l := ORD(code[curMod.codeBase + curFix.pc]); INC(l, curMod.importTab[l].globalBase);
        code[curMod.codeBase + curFix.pc] := CHR(l);
        curFix := curFix.next
      END;
      INC(l0)
    END;
    curMod := curMod.next
  END;
(* Insert a 'LCALL' instruction to the command *)
  IF (result # Ok) THEN RETURN END;
  result := CmdNotFound;  (* 'result' will be set to 'Ok' if the command was found *)
  IF (mainMod # NIL) THEN
    l := 0;
    WHILE (l < mainMod.nofCommands) & (mainMod.commandTab[l].name # cmdName) DO INC(l) END;
    IF (l < mainMod.nofCommands) THEN  (* command found *)
      result := Ok;
      cmdAdr := mainMod.codeBase + mainMod.commandTab[l].entryAdr;
      code[pc] := CHR(XOC.LCALL);  (* call the command *)
      code[pc+1] := CHR(cmdAdr DIV 100H);
      code[pc+2] := CHR(cmdAdr MOD 100H);
      INC(pc, 3);
    END
  END;
(* If a trap procedure is explicitely defined, insert a Trap(1) (command terminated) *)
  IF (trapProc # NIL) THEN
    addr := trapProc.owner.codeBase + trapProc.intProcAdr;
    code[pc] := CHR(XOC.MOVac);  (* 'MOV A,#1' *)
    code[pc+1] := 1X;  (* trap number *)
    code[pc+2] := CHR(XOC.PUSH);  (* 'PUSH ACC' *)
    code[pc+3] := CHR(XOC.ACC);
    code[pc+4] := CHR(XOC.LCALL);  (* 'LCALL addr' *)
    code[pc+5] := CHR(addr DIV 100H);
    code[pc+6] := CHR(addr MOD 100H);
    INC(pc, 7)
  END;
(* Finally, insert an endless loop so the controller remains in a defined state after termination of the main command *)
  code[pc] := CHR(XOC.SJMP); code[pc+1] := 0FEX; INC(pc, 2);  (* 'SJMP -2' (endless loop) *)
(* Fixup all trap calls *)
  curMod := moduleList; addr := -1;
  WHILE (curMod # NIL) DO
    curFix := curMod.trapCallList;
    WHILE (curFix # NIL) DO
      IF (addr = -1) THEN  (* first call of the only trap procedure *)
        IF (trapProc = NIL) THEN  (* insert anonymous trap procedure *)
          NEW(trapProc); trapProc.intProcAdr := pc; trapProc.owner := NIL;
          code[pc] := CHR(XOC.RET); INC(pc); addr := trapProc.intProcAdr
        ELSE addr := trapProc.owner.codeBase + trapProc.intProcAdr
        END
      END;
      code[curMod.codeBase + curFix.pc] := CHR(addr DIV 100H);
      code[curMod.codeBase + curFix.pc + 1] := CHR(addr MOD 100H);
      curFix := curFix.next
    END;
    curMod := curMod.next
  END
END CreateLayout;

PROCEDURE WriteHexFile(name, cmd: ARRAY OF CHAR);  (* generate intel-hex file of addresses 0 to 'pc-1' *)
VAR
  hexFile: XOH.OutputFile;
  adr: LONGINT;
  checkSum, nofBytes: INTEGER;

  PROCEDURE OutHex(val: LONGINT);
  BEGIN checkSum := (checkSum + SHORT(val)) MOD 100H; hexFile.WriteHex8(val)
  END OutHex;

BEGIN
  hexFile := XOH.NewOutputFile(XOH.path, name, XOH.IntelHexFileExtension);
  IF (hexFile = NIL) THEN result := NoOutputFile; RETURN END;
  XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote); XOH.LogWriteString(hexFile.name);
  XOH.LogWrite("."); XOH.LogWriteString(cmd); XOH.LogWrite(XOH.DoubleQuote);
  adr := 0;
  WHILE (adr < pc) DO
    checkSum := 0; hexFile.Write(":");
    IF (pc-adr >= IntelHexNofData) THEN nofBytes := IntelHexNofData  (* full line *)
    ELSE nofBytes := SHORT(pc-adr)
    END;
    OutHex(nofBytes);  (* number of data bytes *)
    OutHex(adr DIV 100H); OutHex(adr MOD 100H);  (* MSB and LSB of the address *)
    OutHex(00H);  (* identification of data line *)
    WHILE (nofBytes > 0) DO
      OutHex(ORD(code[adr])); INC(adr); DEC(nofBytes)
    END;
    OutHex(100H - checkSum);
    hexFile.Write(XOH.CR); hexFile.Write(XOH.LF)
  END;  (* end line: don't use 'WriteString' because the string termination symbol '0X' should not be written *)
  hexFile.Write(":"); hexFile.Write("0"); hexFile.Write("0");
  hexFile.Write("0"); hexFile.Write("0"); hexFile.Write("0"); hexFile.Write("0");
  hexFile.Write("0"); hexFile.Write("1"); hexFile.Write("F"); hexFile.Write("F");
  hexFile.Write(XOH.CR); hexFile.Write(XOH.LF);  (* end line *)
  hexFile.Register; hexFile.Close
END WriteHexFile;

PROCEDURE WriteBinFile(name, cmd: ARRAY OF CHAR);  (* generate binary output file of addresses 0 to 'pc-1' *)
VAR binFile: XOH.OutputFile;
BEGIN
  binFile := XOH.NewOutputFile(XOH.path, name, XOH.BinaryFileExtension);
  IF (binFile = NIL) THEN result := NoOutputFile; RETURN END;
  XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote); XOH.LogWriteString(binFile.name);
  XOH.LogWrite("."); XOH.LogWriteString(cmd); XOH.LogWrite(XOH.DoubleQuote);
  binFile.WriteBytes(code, pc);
  binFile.Register; binFile.Close
END WriteBinFile;

PROCEDURE DisplayInformation;  (* lists the linker information *)
VAR
  mod: Module;
  int: InterruptProcedure;
  i, lastAdr: INTEGER;
BEGIN
  XOH.LogWriteLn; XOH.LogWriteString("  code memory space");
  int := intProcList;
  WHILE (int # NIL) DO
    XOH.LogWriteLn; XOH.LogWriteString("    "); XOH.LogWriteHex16(int.intAdr);
    XOH.LogWriteString("H interrupt served by "); XOH.LogWriteHex16(int.owner.codeBase + int.intProcAdr);
    XOH.LogWriteString("H ("); XOH.LogWriteString(int.owner.name);
    i := 0; WHILE (i < int.owner.nofCommands) & (int.owner.commandTab[i].entryAdr # int.intProcAdr) DO INC(i) END;
    IF (i < int.owner.nofCommands) THEN XOH.LogWrite("."); XOH.LogWriteString(int.owner.commandTab[i].name) END;
    XOH.LogWrite(")");
    int := int.next
  END;
  mod := moduleList;
  WHILE (mod # NIL) DO
    XOH.LogWriteLn; XOH.LogWriteString("    "); XOH.LogWriteHex16(mod.codeBase);
    XOH.LogWriteString("H .. "); XOH.LogWriteHex16(mod.codeBase + mod.codeSize - 1);
    XOH.LogWriteString("H   "); XOH.LogWriteString(mod.name);
    mod := mod.next
  END;
  XOH.LogWriteLn; XOH.LogWriteString("  external RAM");
  XOH.LogWriteLn; XOH.LogWriteString("    ");
  IF (options.extRAMbeg # options.extRAMend) THEN  (* external RAM available *)  
    XOH.LogWriteHex16(options.extRAMbeg); XOH.LogWriteString("H .. "); XOH.LogWriteHex16(options.extRAMend);
    XOH.LogWriteString("H   ("); XOH.LogWriteInt(options.extRAMend - options.extRAMbeg + 1);
    XOH.LogWriteString(" bytes for heap)")
  ELSE XOH.LogWriteString("none")
  END;
  XOH.LogWriteLn; XOH.LogWriteString("  "); XOH.LogWriteString("internal RAM");
  XOH.LogWriteLn; XOH.LogWriteString("  "); XOH.LogWriteString("  "); XOH.LogWriteString("00H .. 07H   registers R0..R7");
  IF (options.extRAMbeg # options.extRAMend) THEN  (* external RAM available *)
    XOH.LogWriteLn; XOH.LogWriteString("  "); XOH.LogWriteString("  "); XOH.LogWriteString("08H .. 09H   heap pointer");
    lastAdr := 0AH
  ELSE lastAdr := 08H
  END;
  mod := moduleList;
  WHILE (mod # NIL) DO
    IF (mod.globalData > 0) THEN
      XOH.LogWriteLn; XOH.LogWriteString("  "); XOH.LogWriteString("  "); XOH.LogWriteHex8(SHORT(mod.globalBase));
      XOH.LogWrite("H");
      IF (mod.globalData > 1) THEN
        XOH.LogWriteString(" .. "); XOH.LogWriteHex8(SHORT(mod.globalBase + mod.globalData - 1)); XOH.LogWrite("H")
      END;
      XOH.LogWriteString("  global"); IF (mod.globalData > 1) THEN XOH.LogWrite("s") END;
      XOH.LogWriteString(" of "); XOH.LogWriteString(mod.name);
      lastAdr := SHORT(mod.globalBase + mod.globalData)
    END;
    mod := mod.next
  END;
  IF (sp - lastAdr + 1 > 0) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  "); XOH.LogWriteString("  "); XOH.LogWriteHex8(lastAdr); XOH.LogWrite("H");
    IF (lastAdr # sp) THEN
      XOH.LogWriteString(" .. "); XOH.LogWriteHex8(SHORT(sp)); XOH.LogWrite("H")
    END;
    XOH.LogWriteString("   (");
    IF (sp - lastAdr + 1 > 1) THEN
      XOH.LogWriteInt(sp - lastAdr + 1); XOH.LogWriteString(" bytes")
    ELSE XOH.LogWriteString("one byte")
    END; XOH.LogWriteString(" for local variabes)")
  END;
  XOH.LogWriteLn; XOH.LogWriteString("  "); XOH.LogWriteString("  ");
  XOH.LogWriteHex8(SHORT(sp + 1));  (* SP will be preincremented *)
  XOH.LogWriteString("H .. "); XOH.LogWriteHex8(SHORT(options.intRAM - 1)); XOH.LogWriteString("H   (");
  XOH.LogWriteInt(options.intRAM - sp - 1); XOH.LogWriteString(" bytes for stack)")
END DisplayInformation;

PROCEDURE Link*;  (* to merge '.obj' files into a '.bin' file *)
VAR
  i, j: INTEGER;
  ch: CHAR;
  mainMod, additionalMod: Module;
  mainModName, cmdName, additionalModName: XOH.Id;

  PROCEDURE SkipBlanks;
  BEGIN WHILE (param[i] # 0X) & ((param[i] = " ") OR (param[i] = XOH.TAB)) DO INC(i) END; ch := param[i]
  END SkipBlanks;

BEGIN
  XOH.LogWriteString("linking");
  moduleList := NIL; intProcList := NIL; trapProc := NIL;  (* empty module list, interrupt procedure list and trap procedure *)
  pc := 0; globalData := 0; maxLocalData := 0; options.valid := FALSE;
  IF (param[0] = 0X) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  invalid name");
    XOH.LogWriteLn; RETURN
  END;
  i := 0; j := 0; SkipBlanks; ch := param[i];
  WHILE (j < XOH.MaxIdLen) & (i < XOH.MaxStrLen) & (ch # 0X) & (ch # ".") DO
    mainModName[j] := ch; INC(j); INC(i); ch := param[i]
  END;
  IF (ch # ".") THEN
    XOH.LogWriteLn; XOH.LogWriteString("  failed ( 'Module.Command {Module}' format expected)");
    XOH.LogWriteLn; RETURN
  END;
  INC(i);
  IF (j >= XOH.MaxIdLen) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  failed (module name too large)");
    XOH.LogWriteLn; RETURN
  END;
  mainModName[j] := 0X; j := 0; ch := param[i];
  WHILE (j < XOH.MaxIdLen) & (i < XOH.MaxStrLen) & (ch # 0X) & (ch # " ") DO
    cmdName[j] := ch; INC(j); INC(i); ch := param[i]
  END;
  IF (j >= XOH.MaxIdLen) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  failed (command name too large)");
    XOH.LogWriteLn; RETURN
  END;
  cmdName[j] := 0X;
  mainMod := ThisModule(mainModName);  (* load the main module and its imported modules *)
  WHILE (ch # 0X) & (result = Ok) DO
    SkipBlanks; j := 0;
    WHILE (j < XOH.MaxIdLen) & (i < XOH.MaxStrLen) & (ch # 0X) & (ch # " ") DO
      additionalModName[j] := ch; INC(j); INC(i); ch := param[i]
    END;
    IF (j >= XOH.MaxIdLen) THEN
      XOH.LogWriteLn; XOH.LogWriteString("  failed (module name too large)")
    ELSE
      additionalModName[j] := 0X; additionalMod := ThisModule(additionalModName)  (* load an additional module *)
    END
  END;
  IF (result = Ok) THEN  (* no error occurred *)
    CreateLayout(mainMod, cmdName);
    IF (result = Ok) THEN
      IF intelHex THEN WriteHexFile(mainModName, cmdName)
      ELSE WriteBinFile(mainModName, cmdName)
      END
    END
  END;
  CASE result OF
  | Ok:
    XOH.LogWriteString("   "); XOH.LogWriteInt(pc);
    XOH.LogWriteString("   "); XOH.LogWriteInt(globalData);
    XOH.LogWriteString("   "); XOH.LogWriteInt(maxLocalData);
    DisplayInformation;
    IF (trapProc # NIL) & (trapProc.owner = NIL) THEN
      (* trap procedure is used, and an anonymous trap procedure has been inserted *)
      XOH.LogWriteLn; XOH.LogWriteString("  warning: no trap procedure defined")
    END
  | ObjNotFound:
    XOH.LogWriteLn; XOH.LogWriteString("  object file from module "); XOH.LogWriteString(XOH.DoubleQuote);
    XOH.LogWriteString(imported); XOH.LogWriteString(XOH.DoubleQuote); XOH.LogWriteString(" not found")
  | BadObjFile: XOH.LogWriteLn; XOH.LogWriteString("  failed (referenced file is not an object file)")
  | BadKey:
    XOH.LogWriteLn; XOH.LogWriteString("  call error: "); XOH.LogWriteString(importing);
    XOH.LogWriteString(" imports "); XOH.LogWriteString(imported); XOH.LogWriteString(" with bad key")
  | OptionsInconsistency: XOH.LogWriteLn; XOH.LogWriteString("  failed (options inconsistency)")
  | MultiplyInterrupt: XOH.LogWriteLn; XOH.LogWriteString("  failed (multiply defined interrupt service procedure)")
  | MultiplyTrap: XOH.LogWriteLn; XOH.LogWriteString("  failed (multiply defined trap procedure)")
  | CmdNotFound: XOH.LogWriteLn; XOH.LogWriteString("  failed (command not found)")
  | CodeTooLong: XOH.LogWriteLn; XOH.LogWriteString("  failed (linked code too long)")
  | NoOutputFile: XOH.LogWriteLn; XOH.LogWriteString("  failed (output file could not be generated)")
  END;
  XOH.LogWriteLn; moduleList := NIL  (* objects are not used any more; garbage collector may collect them *)
END Link;

(* Loader *)

PROCEDURE ReadHexFile(inputFile: XOH.InputFile): LONGINT;  (* returns the highest code address, or '-1' if error *)
VAR
  ch: CHAR;
  i, nofBytes, lineId, checkSum, maxPc: LONGINT;

  PROCEDURE InHex(VAR val: LONGINT);
  BEGIN inputFile.ReadHex8(val); checkSum := (checkSum + val) MOD 100H
  END InHex;

BEGIN
  maxPc := 0;
  lineId := 0H; inputFile.Read(ch);
  WHILE (lineId = 0H) & ~inputFile.Eof() DO
    IF (ch # ":") THEN RETURN -1 END;  (* first character of a line must be a colon *)
    checkSum := 0;
    InHex(nofBytes);
    InHex(i); pc := 100H*i; InHex(i); INC(pc, i);
    InHex(lineId);
    WHILE (nofBytes > 0) DO InHex(i); code[pc] := CHR(i); INC(pc); DEC(nofBytes) END;
    IF (pc-1 > maxPc) THEN maxPc := pc-1 END;
    InHex(i); IF (checkSum # 0H) THEN RETURN -1 END;
    IF (lineId = 0H) THEN  (* data line *)
      inputFile.Read(ch); WHILE (ch # ":") DO inputFile.Read(ch) END  (* skip CR and LF *)
    END
  END;
  RETURN maxPc
END ReadHexFile;

PROCEDURE Load*;  (* copy a '.hex' or a '.bin' file to parallel port *)
VAR
  ch: CHAR;
  inputFile: XOH.InputFile;
  pc, maxPc: LONGINT;
BEGIN
  XOH.LogWriteString("loading");
  IF ~XOH.InitLPT(lptPort) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  parallel port not ready")
  ELSIF intelHex THEN  (* load intel-hex file *)
    inputFile := XOH.SelectInputFile(XOH.path, XOH.IntelHexFileExtension);
    IF (inputFile = NIL) THEN XOH.LogWriteLn; XOH.LogWriteString("  file not found")
    ELSE
      XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote);
      XOH.LogWriteString(inputFile.name); XOH.LogWrite(XOH.DoubleQuote);
      maxPc := ReadHexFile(inputFile);
      IF (maxPc # -1) THEN
        FOR pc := 0 TO maxPc DO XOH.OutLPT(lptPort, code[pc]) END;
        XOH.LogWriteString("   "); XOH.LogWriteInt(pc)  (* 'pc' is = 'maxPc'+1 = number of loaded bytes *)
      ELSE
        XOH.LogWriteLn; XOH.LogWriteString("  corrupt hex file")
      END; inputFile.Close
    END
  ELSE  (* load binary file *)
    inputFile := XOH.SelectInputFile(XOH.path, XOH.BinaryFileExtension);
    IF (inputFile = NIL) THEN XOH.LogWriteLn; XOH.LogWriteString("  file not found")
    ELSE
      XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote);
      XOH.LogWriteString(inputFile.name); XOH.LogWrite(XOH.DoubleQuote);
      inputFile.Read(ch); pc := 0;
      WHILE ~inputFile.Eof() DO XOH.OutLPT(lptPort, ch); INC(pc); inputFile.Read(ch) END;
      XOH.LogWriteString("   "); XOH.LogWriteInt(pc); inputFile.Close
    END
  END; XOH.LogWriteLn
END Load;

END Mcs51XOL.

